/*
 * Copyright (c) 2010 SURFnet bv
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*****************************************************************************
 HashAlgorithm.cpp

 Base class for hash algorithm classes
 *****************************************************************************/

#include "config.h"
#include "HashAlgorithm.h"

// Base constructor
HashAlgorithm::HashAlgorithm()
{
	currentOperation = NONE;
}

// Hashing functions
bool HashAlgorithm::hashInit()
{
	if (currentOperation != NONE)
	{
		return false;
	}

	currentOperation = HASHING;

	return true;
}

bool HashAlgorithm::hashUpdate(const ByteString& /*data*/)
{
	if (currentOperation != HASHING)
	{
		return false;
	}

	return true;
}

bool HashAlgorithm::hashFinal(ByteString& /*hashedData*/)
{
	if (currentOperation != HASHING)
	{
		return false;
	}

	currentOperation = NONE;

	return true;
}

